/****************************************************************************
 *
 *	File		: arpdata.c
 *
 *	Title		: ARP Protocol Data Engine
 *
 *	Author		: Andrae Muys
 *
 *	Revisions	: 2nd Nov 1999	AWM - Initial release.
 *
 ****************************************************************************/

#include <gtk/gtk.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <byteswap.h>
#include <string.h>
#include <stdio.h>

#include "arpdata.h"
#include "snif.h"
#include "show.h"

static arpData *arpTableLookup (ARPTable *table, u_int8_t *ip, u_int8_t size);
static void arpTableUpdateIP (arpData *data, u_int32_t ip);
static guint arpHashFunc (gconstpointer key);
static gint arpCompareFunc (gconstpointer a, gconstpointer b);
static gint arpIPCompareFunc (gconstpointer a, gconstpointer b);

/****** EXTERNAL FUNCTIONS ******/

ARPTable *arpTableInit ()
{
	ARPTable *table;

	table = g_malloc (sizeof(ARPTable));
	table->table = g_hash_table_new (arpHashFunc, arpCompareFunc);
	table->chunk = g_mem_chunk_new ("ARPTable", sizeof(arpData), 4096,
			G_ALLOC_ONLY);
	table->ip = g_mem_chunk_new ("arpIP", sizeof(arpIP), 4096,
			G_ALLOC_AND_FREE);

	return table;
}

void arpTableUpdate (Snif *snif)
{
	arpData *src;
	arpData *dest;

	struct arphdr *hdr = (struct arphdr *)snif->nethdr;

	if (g_ntohs (hdr->ar_pro) != ETHERTYPE_IP) {
		return;
	}

	switch (g_ntohs (hdr->ar_op)) {
	case ARPOP_REPLY :
		dest = arpTableLookup (arpTable, 
				(snif->data + hdr->ar_hln + hdr->ar_pln), hdr->ar_hln);
		arpTableUpdateIP (dest, 
			g_ntohl ((guint)(snif->data + (2 * hdr->ar_hln) + hdr->ar_pln)));
	/* Debugging...
		showARPUpdate (dest);
	*/
	case ARPOP_REQUEST :	/* Intentional fallthrough */
		src = arpTableLookup (arpTable, snif->data, hdr->ar_hln);
		arpTableUpdateIP (src, g_ntohl ((guint)(snif->data + hdr->ar_hln)));
	/*
		showARPUpdate (src);
	*/
		break;
	case ARPOP_RREQUEST :
		/* Not supported yet */
	case ARPOP_RREPLY :
		/* Not supported yet */
	default :
		return;
	}

	return;
}

/****** STATIC FUNCTIONS ******/

static arpData *arpTableLookup (ARPTable *table, u_int8_t *haddr, u_int8_t size)
{
	arpData *data;

	data = (arpData *)g_hash_table_lookup (table->table, haddr);

	if (!data) {
		data = g_mem_chunk_alloc (table->chunk);
		data->table = table;
		data->len = size;
		memcpy (data->hwaddr, haddr, size);
		memset (data->hwaddr + size, 0, MAX_ADDR_LEN - size);
		data->ip = NULL;
		data->node = NULL;

		g_hash_table_insert (table->table, haddr, data);
	}

	return data;
}

#define rawtoETHstr(addr) "%02x:%02x:%02x:%02x:%02x:%02x", \
	((char *)&(addr))[0], ((char *)&(addr))[1], ((char *)&(addr))[2], \
	((char *)&(addr))[3], ((char *)&(addr))[4], ((char *)&(addr))[5] 

static void arpTableUpdateIP (arpData *data, u_int32_t ip)
{
	arpIP *ipdata;
	GList *node;
	char iplabel[16];

	ipdata = g_mem_chunk_alloc (data->table->ip);
	ipdata->ip = ip;
	node = g_list_find_custom (data->ip, ipdata, arpIPCompareFunc);
	if (!node) {
		g_snprintf (iplabel, 19, rawtoETHstr (data->hwaddr));
		ipdata->node = gtk_tree_item_new_with_label (iplabel);
	} else {
		g_mem_chunk_free (data->table->ip, ipdata);
		data->ip = g_list_remove_link (data->ip, node);
		ipdata = node->data;
	}

	data->ip = g_list_prepend (data->ip, ipdata);
	gtk_object_ref (GTK_OBJECT(ipdata->node));
	data->update = g_list_prepend (data->update, ipdata->node);
}

static guint arpHashFunc (gconstpointer key)
{
	return g_ntohl ((guint)key);
}

static gint arpCompareFunc (gconstpointer a, gconstpointer b)
{
	return memcmp (a, b, MAX_ADDR_LEN);
}

static gint arpIPCompareFunc (gconstpointer a, gconstpointer b)
{
	arpIP *aptr = (arpIP *)a;
	arpIP *bptr = (arpIP *)b;

	return (aptr->ip == bptr->ip ? 0 : aptr->ip < bptr->ip ? -1 : +1);
}
