/****************************************************************************
 *
 *	File		: packet.c
 *
 *	Title		: Packet Analysis and Processing.
 *
 *	Description	: This file parses the packets and as the exisitance of each
 *					protocol is determined it is passed to the appropriate
 *					protocol data engine for final processing.  Be aware that
 *					the buffer passed by the capture engine is handled by the
 *					capture engine and should not be deallocated, or modified in
 *					anyway packet.c or anyother file.
 *
 *	Author		: Andrae Muys
 *
 *	Revisions	: 2nd Nov 1999  AWM - Initial release.
 *
 ****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h> 
#include <byteswap.h>

#include <gtk/gtk.h>

#include "snif.h"
#include "packet.h"
#include "tcpdata.h"
#include "ipdata.h"
#include "udpdata.h"
#include "arpdata.h"

#define poiu fprintf (stderr, "Checkpoint %s:%d\n", __FILE__, __LINE__)

void handle_packet (gpointer data, gint source, GdkInputCondition cond)
{
	Snif *snif = (Snif *)data;

	snif->len = read (snif->sd, snif->buf, snif->mtu);
	if (snif->len < 0) {
		perror ("pasmon");
		return;
	}

	switch (snif->llproto) {
	case ETHERNET : /* IP */
		snif->llhdr = snif->buf;
		handle_ethernet_packet (snif);
		break;
	default : /* Unknown */
		fprintf (stderr, "Error.  Unknown Link Level Protocol.\n");
		return;
	}
}

void handle_ethernet_packet (Snif *snif)
{
	struct ether_header *ethhdr;

	ethhdr = (struct ether_header *)snif->llhdr;

	switch (g_ntohs (ethhdr->ether_type)) {
	case ETHERTYPE_IP :
		snif->nethdr = snif->llhdr + 14;
		handle_ip_packet (snif);
		break;
	case ETHERTYPE_ARP :
	/*
		snif->nethdr = snif->llhdr + 14;
		handle_arp_packet (snif);
		break;
	case ETHERTYPE_REVARP :
		snif->nethdr = snif->llhdr + 14;
		handle_rarp_packet (snif);
	*/
		break;
	default : /* Probably IEEE 802.2/802.3 Encapsulation */
		fprintf (stderr, "Unknown Ethernet type detected. Should try IEEE.\n");
	#if 0
		/*
		 * I don't really care if this works.  IEEE Encap' sucks.  It isn't
		 * unabiguously differentiable from RFC894 Encap' as the Length
		 * field can clash with Xerox PUP.  Simply better if it didn't
		 * exist.  Anyway, hopefully this works, if it dosn't fix it
		 * yourself if you need it.
		 */
		switch ((u_int16_t)(snif->llhdr[20])) {
		case ETHERTYPE_IP :
			snif->nethdr = snif->llhdr + 22;
			handle_ip_packet (snif);
			break;
		case ETHERTYPE_ARP :
			snif->nethdr = snif->llhdr + 22;
			handle_arp_packet (snif);
			break;
		case ETHERTYPE_REVARP :
			snif->nethdr = snif->llhdr + 22;
			handle_rarp_packet (snif);
			break;
		}
	#endif
	}
}

void handle_arp_packet (Snif *snif)
{
	snif->data = snif->nethdr + 8;

	arpTableUpdate (snif);
}

void handle_rarp_packet (Snif *snif)
{
	/* RARP not handled yet */
}

void handle_ip_packet (Snif *snif)
{
	struct iphdr *hdr = (struct iphdr *)snif->nethdr;

	snif->tpthdr = snif->nethdr + (hdr->ihl * 4);
	snif->end = snif->nethdr + hdr->tot_len;

	ipTableUpdate (snif);

	switch (hdr->protocol) {
	case IPPROTO_TCP :
		handle_tcp_packet (snif);
		break;
	case IPPROTO_UDP :
		handle_udp_packet (snif);
		break;
	}
}

void handle_tcp_packet (Snif *snif)
{
	struct iphdr *iph = (struct iphdr *)snif->nethdr;
	struct tcphdr *hdr = (struct tcphdr *)snif->tpthdr;

	snif->data = snif->tpthdr + (hdr->doff * 4);
	if (iph->version == 5) {
		printf ("Hello\n");
	}
	connTableUpdate (snif);
}

void handle_udp_packet (Snif *snif)
{
	snif->data = snif->tpthdr + 8;

	udpTableUpdate (snif);
}
