/****************************************************************************
 *
 *  File        : tcpdata.c
 *
 *  Title       : TCP Protocol Data Engine
 *
 *  Author      : Andrae Muys
 *
 *  Revisions   : 2nd Nov 1999  AWM - Initial release.
 *
 ****************************************************************************/


#include <gtk/gtk.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <net/ethernet.h>
#include <byteswap.h>
#include <string.h>
#include <time.h>
#include <stdio.h>

#include "tcpdata.h"
#include "snif.h"
#include "show.h"

char *tcpStateStr[NTCPSTATES] = { "SYN", "ESTABLISHED", "FINISHED", "CLOSED" };

static gint connTableCompareFunc (gconstpointer a, gconstpointer b);
static guint connTableHashFunc (gconstpointer a);
static connData *connTableLookup (ConnTable *table, 
									u_int32_t ip1, u_int16_t port1, 
									u_int32_t ip2, u_int16_t port2);
static void move_to_closed_list (connData *conn);

/****** EXTERNAL FUNCTIONS ******/

ConnTable *connTableInit ()
{
	ConnTable *table;

	table = g_malloc (sizeof(ConnTable));
	table->table = g_hash_table_new (connTableHashFunc, connTableCompareFunc);
	table->chunk = g_mem_chunk_new ("ConnTable", sizeof(connData), 4096,
			G_ALLOC_ONLY);
	table->key = g_mem_chunk_new ("Index ConnTable", sizeof(connIndex), 4096,
			G_ALLOC_ONLY);

	return table;
}

void connTableUpdate (Snif *snif)
{
	connData *conn;
	struct iphdr *iph = (struct iphdr *)snif->nethdr;
	struct tcphdr *tcph = (struct tcphdr *)snif->tpthdr;

	conn = connTableLookup (connTable, bswap_32 (iph->saddr), 
			bswap_16 (tcph->source), bswap_32 (iph->daddr),
			bswap_16 (tcph->dest));
	
	conn->volume += bswap_16 (iph->tot_len);
	conn->last = time (NULL);

	if (conn->state == SYN && !tcph->syn) {
		conn->state = ESTAB;
	}
	if (conn->state == ESTAB && tcph->fin) {
		conn->state = FIN;
	/*
	 * I'm not sure how to handle half-closes in this passive environment.  The
	 * problem is that it is possible for me to miss packets, therefore
	 * missing the normal case of FIN/FIN-ACK full close.  This is complicated
	 * by the fact that after 2MSL (anywhere from 1-4 minutes, and independent
	 * of pasmon's host) the socket-pair I'm using to index the connTable can be
	 * reused, so data appearing on a previously FIN'd connection could be the
	 * result of a half-close, or a reused socket-pair.  Worse still, BSD
	 * stacks can permit a reused socket-pair within the 2MSL RFC timeframe.
	 *
	 * I've decided to assume that half-closes don't happen, and therefore that
	 * any further traffic is the result of a new connection on a reused
	 * socket-pair.  I will however await the first ACK'd FIN (with a timeout to
	 * avoid the hassle should I miss the ACK) before considering the connection
	 * closed.
	 */
	} else if (conn->state == FIN && tcph->ack) {
		conn->state = CLOSED;
		move_to_closed_list (conn);
	}

	showConnUpdate (conn);
};

/****** STATIC FUNCTIONS ******/

static connData *connTableLookup (ConnTable *table, 
									u_int32_t ip1, u_int16_t port1, 
									u_int32_t ip2, u_int16_t port2)
{
	connIndex *key;
	connData *data;

	key = g_mem_chunk_alloc (table->key);
	key->ip1 = ip1;
	key->port1 = port1;
	key->ip2 = ip2;
	key->port2 = port2;

	data = (connData *)g_hash_table_lookup (table->table, key);

	if (!data) {
		data = g_mem_chunk_alloc (table->chunk);
		data->hdr = key;
		data->dup = NULL;
		data->volume = 0;
		data->start = time (NULL);
		data->last = data->start;
		data->state = SYN;
		data->row = NULL;

		g_hash_table_insert (table->table, key, data);
	}

	return data;
}

static void move_to_closed_list (connData *conn)
{
	/* Move away from memory */
}

static gint connTableCompareFunc (gconstpointer a, gconstpointer b)
{
	connIndex *aIndex = (connIndex *)a;
	connIndex *bIndex = (connIndex *)b;

	/*
	 * Either A(ip1,port,ip2,port) == B(ip1,port,ip2,port) or
	 * 									B(ip2,port,ip1,port).
	 */

	return ((aIndex->ip1 == bIndex->ip1 &&
				aIndex->port1 == bIndex->port1 &&
				aIndex->ip2 == bIndex->ip2 &&
				aIndex->port2 == bIndex->port2) ||
			(aIndex->ip1 == bIndex->ip2 &&
				aIndex->port1 == bIndex->port2 &&
				aIndex->ip2 == bIndex->ip1 &&
				aIndex->port2 == bIndex->port1));
}

static guint connTableHashFunc (gconstpointer a)
{
	connIndex *aIndex = (connIndex *)a;

	return ((uint)aIndex->ip1 ^ (uint)aIndex->port1 ^ 
			(uint)aIndex->ip2 ^ (uint)aIndex->port2);
}

